/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <tizen.h>
#include <Elementary.h>
#include <system_settings.h>
#include <badge.h>
#include "$(appName).h"
#include "view.h"

static bool _update_badge_request_cb(const pkginfo_s *pkginfo, int badge_count);
static void _get_badge_request_cb(pkginfo_s **pkginfo);
static unsigned int _get_badge_count(const pkginfo_s *pkginfo);
static bool _create_new_badge(const char *app_id, unsigned int badge_count);
static bool _remove_badge(const char *app_id);
static bool _update_badge(const char *app_id, unsigned int badge_count);

/**
 * @brief Hook to take necessary actions before main event loop starts.
 * Initialize UI resources and application's data.
 * If this function returns true, the main loop of application starts.
 * If this function returns false, the application is terminated.
 */
static bool app_create(void *user_data)
{
	int i;
	int apps_count;
	pkginfo_s *pkginfo;

	data_initialize();

	view_set_callbacks(_update_badge_request_cb, _get_badge_request_cb);

	if (!view_create(NULL))
		return false;

	apps_count = data_get_application_count();

	for (i = 0; i < apps_count; i++)
		if (data_get_application_info(i, &pkginfo) && pkginfo->changeable)
			view_add_application(pkginfo);

	for (i = 0; i < apps_count; i++)
		if (data_get_application_info(i, &pkginfo) && !pkginfo->changeable)
			view_add_application(pkginfo);

	return true;
}

/**
 * @brief This callback function is called when another application
 * sends a launch request to the application.
 */
static void app_control(app_control_h app_control, void *user_data)
{
	/* Handle the launch request. */
}

/**
 * @brief This callback function is called each time
 * the application is completely obscured by another application
 * and becomes invisible to the user.
 */
static void app_pause(void *user_data)
{
	/* Take necessary actions when application becomes invisible. */
}

/**
 * @brief This callback function is called each time
 * the application becomes visible to the user.
 */
static void app_resume(void *user_data)
{
	/* Take necessary actions when application becomes visible. */
}

/**
 * @brief This callback function is called once after the main loop of the application exits.
 */
static void app_terminate(void *user_data)
{
	/* Release all resources. */
	view_destroy();
	data_finalize();
}

/**
 * @brief This function will be called when the language is changed.
 */
static void ui_app_lang_changed(app_event_info_h event_info, void *user_data)
{
	/* APP_EVENT_LANGUAGE_CHANGED */
	char *locale = NULL;

	system_settings_get_value_string(SYSTEM_SETTINGS_KEY_LOCALE_LANGUAGE, &locale);

	if (locale != NULL) {
		elm_language_set(locale);
		free(locale);
	}
	return;
}

/**
 * @brief Main function of the application.
 */
int main(int argc, char *argv[])
{
	int ret;

	ui_app_lifecycle_callback_s event_callback = {0, };
	app_event_handler_h handlers[5] = {NULL, };

	event_callback.create = app_create;
	event_callback.terminate = app_terminate;
	event_callback.pause = app_pause;
	event_callback.resume = app_resume;
	event_callback.app_control = app_control;

	/*
	 * If you want to handle more events,
	 * please check the application lifecycle guide.
	 */
	ui_app_add_event_handler(&handlers[APP_EVENT_LANGUAGE_CHANGED], APP_EVENT_LANGUAGE_CHANGED, ui_app_lang_changed, NULL);

	ret = ui_app_main(argc, argv, &event_callback, NULL);
	if (ret != APP_ERROR_NONE)
		dlog_print(DLOG_ERROR, LOG_TAG, "ui_app_main() failed. err = %d", ret);

	return ret;
}

/**
 * @brief Internal callback function invoked by the view module
 * each time the badge of an external application needs to be updated.
 * The update method is chosen based on the current value of a badge's counter
 * and a new value to be set to the badge:
 * - if badge counter's current value is 0 and a new value to be set is non-zero, then new badge is created;
 * - if badge counter's current value and a new value to be set are non-zero and are different from each other, then current value is altered with the new one;
 * - if badge counter's current value is non-zero and a new value to be set is 0, then the badge is removed;
 * - if badge counter's current value is equal to the new value to be set, then nothing happens.
 * @param[in] pkginfo The structure containing the application related information.
 * @param[in] badge_count The value to be displayed in the badge of an external application.
 * @return This function returns 'true' is the application's badge was updated successfully,
 * otherwise 'false' is returned.
 */
static bool _update_badge_request_cb(const pkginfo_s *pkginfo, int badge_count)
{
	bool ret;
	int curr_badge_count = _get_badge_count(pkginfo);

	if (badge_count == curr_badge_count) {
		dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': no changes introduced.", pkginfo->app_id);
		return true;
	}

	if (curr_badge_count == 0) {
		ret = _create_new_badge(pkginfo->app_id, badge_count);
		if (ret)
			dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge created with value of %d.", pkginfo->app_id, badge_count);
		else
			dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge creation error.", pkginfo->app_id);

		return ret;
	}

	if (badge_count == 0) {
		ret = _remove_badge(pkginfo->app_id);
		if (ret)
			dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge removed and set to 0.", pkginfo->app_id);
		else
			dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge removal error.", pkginfo->app_id);

		return ret;
	}

	ret = _update_badge(pkginfo->app_id, badge_count);
	if (ret)
		dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge counter changed to %d.", pkginfo->app_id, badge_count);
	else
		dlog_print(DLOG_INFO, LOG_TAG, "Badge update for '%s': badge counter change error.", pkginfo->app_id);

	return ret;
}

/**
 * @brief Internal callback function invoked by the view module
 * each time the badge counter of an external application needs to be acquired.
 * @param[in] pkginfo The structure containing the application related information.
 */
static void _get_badge_request_cb(pkginfo_s **pkginfo)
{
	(*pkginfo)->badge_count = _get_badge_count(*pkginfo);
}

/**
 * @brief Internal function responsible for badge counter value acquisition from an external application.
 * @param[in] pkginfo The structure containing the application related information.
 * @return The obtained badge counter value is returned. On error, the function returns 0.
 */
static unsigned int _get_badge_count(const pkginfo_s *pkginfo)
{
	int ret;
	unsigned int badge_count = 0;

	if (!pkginfo->changeable)
		return 0;

	ret = badge_get_count(pkginfo->app_id, &badge_count);
	if (ret == BADGE_ERROR_NOT_EXIST) {
		dlog_print(DLOG_WARN, LOG_TAG, "Badge for '%s' doest not exist.", pkginfo->app_id);
		return 0;
	} else if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "badge_get_count() failed with error '%s' for %s.", get_error_message(ret), pkginfo->app_id);
		return 0;
	}

	dlog_print(DLOG_INFO, LOG_TAG, "Badge count for %s: %u.", pkginfo->app_id, badge_count);

	return badge_count;
}

/**
 * @brief Internal function responsible for badge creation and display for an external application.
 * The newly created badge is initialized with non-zero value.
 * @param[in] app_id The identifier of an external application for which the badge is to be created.
 * @param[in] badge_count The value to be displayed in the area of an external application's badge.
 * @return Function returns 'true' if the badge was created successfully,
 * otherwise 'false' is returned.
 */
static bool _create_new_badge(const char *app_id, unsigned int badge_count)
{
	int ret;

	ret = badge_add(app_id);
	if (ret == BADGE_ERROR_ALREADY_EXIST) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function badge_add() failed because badge already exists.");
	} else if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_add() failed with error %d.", ret);
		return false;
	}

	ret = badge_set_display(app_id, 1);
	if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_set_display() failed with error %d.", ret);
		return false;
	}

	ret = badge_set_count(app_id, badge_count);
	if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_set_count() failed with error %d.", ret);
		return false;
	}

	return true;
}

/**
 * @brief Internal function responsible for external application's badge removal.
 * Before a badge is removed, its counter is set to 0 and its display flag is set to 0 (hide).
 * @param[in] app_id The identifier of an external application which badge is to be removed.
 * @return Function returns 'true' if the badge was removed successfully,
 * otherwise 'false' is returned.
 */
static bool _remove_badge(const char *app_id)
{
	int ret = badge_set_count(app_id, 0);
	if (ret != BADGE_ERROR_NONE)
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_set_count() failed with error %d.", ret);

	ret = badge_set_display(app_id, 0);
	if (ret != BADGE_ERROR_NONE)
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_set_display() failed with error %d.", ret);

	ret = badge_remove(app_id);
	if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_remove() failed with error %d.", ret);
		return false;
	}

	return true;
}

/**
 * @brief Internal function responsible for external application's badge counter update.
 * Before a badge is updated, its display flag is checked against non-zero value,
 * because only visible badges may be modified.
 * @param[in] app_id The identifier of an external application which badge is to be updated.
 * @param[in] badge_count The value to be set to the badge of an external application.
 * @return Function returns 'true' if the badge was updated successfully,
 * otherwise 'false' is returned.
 */
static bool _update_badge(const char *app_id, unsigned int badge_count)
{
	unsigned int is_display = 0;

	int ret = badge_get_display(app_id, &is_display);
	if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_get_display() failed with error %d.", ret);
		return false;
	}

	if (!is_display)
		return false;

	ret = badge_set_count(app_id, badge_count);
	if (ret != BADGE_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function badge_set_count() failed with error %d.", ret);
		return false;
	}

	return true;
}
